/*
2022-05-15
Michael Nowak
http://www.mino-elektronik.de
All information is subject to change!
*/

#define USART0_PICO
#include <rp2040.h>
#include <stdio.h>
#include "pico_fmeter.h"

extern uint32_t SystemCoreClock;

void set_baudrate0(uint32_t baud)
{
  baud = (SystemCoreClock << 2) / baud;
  UART0->UARTIBRD = baud / 64;
  UART0->UARTFBRD = baud % 64;
  UART0->UARTLCR_H = 3 << UART0_UARTLCR_H_WLEN_Pos | 
                     UART0_UARTLCR_H_FEN_Msk;    // RxD+TxD fifo
}

// TX = GPIO16
// RX = GPIO17
void init_uart0(void)
{
  volatile int delay = 10000;
  RESETS_CLR->RESET = RESETS_RESET_uart0_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_uart0_Msk));

  IO_BANK0->GPIO16_CTRL = IO_BANK0_GPIO16_CTRL_FUNCSEL_uart0_tx;
  IO_BANK0->GPIO17_CTRL = IO_BANK0_GPIO17_CTRL_FUNCSEL_uart0_rx;

// test GPIO18: reversed signals if '0'
  IO_BANK0->GPIO18_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  PADS_BANK0_SET->GPIO18 = PADS_BANK0_GPIO0_PUE_Msk;  // pullup = 1
  PADS_BANK0_CLR->GPIO18 = PADS_BANK0_GPIO0_PDE_Msk;  // pulldown = 0
  while(delay--);                                     // setup time for IO-pin
  if(!(SIO->GPIO_IN & 1 << 18)) {                     //  Txd+RxD reversed
  IO_BANK0->GPIO16_CTRL = IO_BANK0_GPIO16_CTRL_FUNCSEL_uart0_tx |
                         1 << IO_BANK0_GPIO0_CTRL_OUTOVER_Pos;
  IO_BANK0->GPIO17_CTRL = IO_BANK0_GPIO17_CTRL_FUNCSEL_uart0_rx |
                         1 <<  IO_BANK0_GPIO1_CTRL_INOVER_Pos;
  }
  
  set_baudrate0(DEF_BAUDRATE);                        // set baudrate 
  UART0_SET->UARTCR = UART0_UARTCR_UARTEN_Msk;        // enable UART0
}	

void uart0_putc(int zeichen)
{
  while (UART0->UARTFR & UART0_UARTFR_TXFF_Msk);
  UART0->UARTDR = zeichen;
}

int putchar(int x)                      // stdout for printf()
{
  if((uint8_t)x == 10) uart0_putc(13);  // LF -> CR+LF
  uart0_putc(x);
  return(x);
}

int getchar(void)
{
  if(UART0->UARTFR & UART0_UARTFR_RXFE_Msk) {
    return(EOF);                        // fifo is emty return -1
  } else {  
    return(UART0->UARTDR & 0xff);       // return lower 8 bits
  }
}  

void sende_str0(char *s)                // ser. output with CR+LF termination
{
  while(*s) uart0_putc(*(s++));
  uart0_putc(13);                       // CR
  uart0_putc(10);                       // LF
} 
